# -*- coding: utf-8 -*-
###############################################################################
#   Copyright (c) 2021 Gaston Alberto Bertolani
#   License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
###############################################################################

import click

from .utils.remove_background import BackgroundEliminator
from .utils.separator import Separator
from .utils.yolo_generator import Stage


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def aicv():
    """
    AI CV UTILITIES
    """
    pass


@aicv.command(name='rmbg', context_settings=CONTEXT_SETTINGS,
              short_help='Remove Backgraund of image')
@click.argument('source', required=True, type=click.Path())
@click.argument('target', required=True, type=click.Path())
@click.option('-r', '--no-origin', type=bool,
              help='Remove old images after finish')
@click.option('-g', '--with-groups', type=bool,
              help='Create category folder of image in target folder')
def rmbg(source, target, no_origin=False, with_groups=False):
    """
    Remove Background of image

    SOURCE: Source of image/s (Folder or Image)

    TARGET: Target of image/s (Folder or Image)
    """
    rmb = BackgroundEliminator(
        source,
        save_path=target,
        no_origin=no_origin,
        with_groups=with_groups,
    )
    rmb.remove_background()
    return True


@aicv.command(name='split', context_settings=CONTEXT_SETTINGS,
              short_help='Split files in folders')
@click.argument('source_path', required=True, type=click.Path())
@click.option('-t', '--train', type=click.Path(),
              required=True, help='Path to the "train" folder')
@click.option('-v', '--validation', type=click.Path(),
              required=True, help='Path to the "validation" folder')
@click.option('-x', '--test', type=click.Path(),
              required=True, help='Path to the "test" folder')
@click.option('-T', '--train-perc', type=click.Path(),
              required=False, help='Percent of train folder',
              default=0.0)
@click.option('-V', '--validation-perc', type=click.Path(),
              required=False, help='Percent of validation folder',
              default=0.0)
@click.option('-X', '--test-perc', type=click.Path(),
              required=False, help='Percent of test folder',
              default=0.0)
@click.option('-f', '--format', type=str,
              required=False, help='Format to find files e.g jpg,png,jepg')
@click.option('-r', '--remove', type=bool, default=False,
              required=False, help='Remove Source Files')
def split(source_path, train_path, validation_path, test_path,
          train_perc, validation_perc, test_perc,
          format, remove):
    """
    Searches for files in a certain folder and
    separates them into different folders according
    to a certain percentage

    SOURCE: Source of image/s (Folder or Image)

    """
    sep = Separator(
        source_path, train_path,
        validation_path=validation_path,
        test_path=test_path,
        train_perc=train_perc,
        validation_perc=validation_perc,
        test_perc=test_perc,
    )
    sep.split_samples(
        format=format,
        remove_source=remove,
    )
    return True


@aicv.command(name='yologen', context_settings=CONTEXT_SETTINGS,
              short_help='Generate Stages from images to train')
@click.option('-b', '--background', type=click.Path(),
              required=True, help='Background Path')
@click.option('-s', '--samples', type=click.Path(),
              required=True, help='Path to the "sample" folder')
@click.option('-o', '--output', type=click.Path(),
              required=True, help='Output folder to save images')
@click.option('-q', '--qty', type=int,
              default=1, help='Quantity of stages')
@click.option('-r', '--rgx', type=str,
              help='Regex to get class from image file name')
@click.option('-f', '--format', type=str,
              required=False, help='Format of images',
              default=1.0)
@click.option('-z', '--sample-size', type=float, default=1.0,
              help='Size of samples to paste in background')
@click.option('-d', '--sample-degree', type=int, default=0,
              help='Angle to rotate Sample')
def yologen(background, samples, output, qty, rgx,
            format, sample_size, sample_degree):
    """
    Generate Stages from images to train
    """
    stage = Stage(
        background, samples, output,
        class_rgx=rgx,
        format=format,
        sample_perc_size=sample_size,
        sample_degree=sample_degree,
    )
    stage.generate_random_scenes(num_frames=qty)
    return True


if __name__ == '__main__':
    aicv()



# -*- coding: utf-8 -*-
###############################################################################
#   Copyright (c) 2021 Gaston Alberto Bertolani
#   License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
###############################################################################

import random
from PIL import Image
from glob import glob
from pathlib import Path
import os
import re
import math

import click


class Sample(object):

    def __init__(self, class_id, image, x, y,
                 bg_w, bg_h, max_classes):
        self.idx = class_id
        self.image = image
        self.width, self.height = image.size
        self.x = x
        self.y = y
        self.lx = self._fix_measure(x, bg_w)
        self.rx = self._fix_measure(x + self.width, bg_w)
        self.ly = self._fix_measure(y, bg_h)
        self.ry = self._fix_measure(y + self.height, bg_w)
        self.center_x = (self.lx + self.rx) // 2
        self.center_y = (self.ly + self.ry) // 2
        self.bg_w = bg_w
        self.bg_h = bg_h
        self.max_classes = max_classes

    def _fix_measure(self, pos, bg_measure):
        """
        Returns the necessary position so that
        it is always inside the background
        """
        if pos <= 0:
            pos = 0
        elif pos >= bg_measure:
            pos = bg_measure
        return pos

    def _truncate(self, number):
        """
        Returns a value truncated to a specific
        number of decimal places.
        """
        factor = 10.0 ** 6
        number += 0.0000001  # Sum lost decimals
        return str(math.trunc(number * factor) / factor).ljust(8, '0')

    def get_yolo_format(self):
        return [
            str(self.idx).zfill(len(str(self.max_classes))),
            self._truncate((self.lx + self.rx)/2/self.bg_w),
            self._truncate((self.ly + self.ry)/2/self.bg_h),
            self._truncate((self.rx - self.lx)/self.bg_w),
            self._truncate((self.ry - self.ly)/self.bg_h),
        ]


class Stage(object):

    def __init__(self, bg_path, input_path, output_path,
                 class_rgx='', format='*.png',
                 sample_perc_size=1.0, sample_degree=0):
        """ Init Stage

        @param bg_path: Path to background to render samples

        @param input_path: Folder Path to find samples.
                            Can be splitted in many categories
                            with different subfolders names

        @param output_path: Folder Path to save images.

        @param class_rgx: Regex to match different categories
                          when there are multiple
                          samples in same folder

        @param sample_perc_size: Size of sample to render (normalized)

        @param sample_degree: Angle to rotate sample (0 to angle)

        @param format: format of images to locate
        """
        self.background_path = bg_path
        self.sample_folder = input_path
        self.output_path = output_path
        self.class_rgx = class_rgx
        self.format = format
        self.sample_perc_size = sample_perc_size
        self.sample_degree = sample_degree
        if '*.' not in self.format:
            self.format = '*.' + self.format
        self.class_folders = {}
        self.class_samples = {}
        if self.class_rgx:
            self._load_samples_by_rgx()
        else:
            self._load_samples_by_folders()

        # Check samples
        self.enable_generator = False
        click.echo("******   Checking Samples  *****")
        for cls_name, sample_lst in self.class_samples.items():
            if not len(sample_lst):
                click.echo("There are not samples in class name %s" % cls_name)
            else:
                self.enable_generator = True
        click.echo("******   Finish Checking  *****")

        # Save and wrie class.obj
        self.class_idx = {}
        class_obj_path = os.path.join(self.output_path, 'class.obj.txt')
        f = open(class_obj_path, 'w+')
        for i, class_name in enumerate(self.class_samples.keys()):
            if class_name not in self.class_idx:
                self.class_idx[class_name] = i
                f.write(class_name + '\n')
        f.close()

    def _load_samples_by_rgx(self):
        """
        Search in folder path all samples
        """
        for sample_path in glob(self.folder_path + '/' + self.format):
            spath = Path(sample_path)
            rgx = re.search(self.class_rgx, spath.suffix)
            if not rgx:
                continue
            class_name = ','.join(rgx.groups())
            if class_name not in self.class_samples:
                self.class_samples[class_name] = []
            self.class_samples[class_name].append(sample_path)
        return True

    def _load_samples_by_folders(self):
        """
        Search samples in subfolders and
        create class name with folder's name
        """
        self._cfolders = glob(self.sample_folder + '/*/')
        for folder_path in self._cfolders:
            class_name = os.path.basename(
                os.path.normpath(folder_path)
            )
            self.class_folders[class_name] = folder_path
            samples_path_lst = glob(folder_path + '/' + self.format)
            self.class_samples[class_name] = samples_path_lst
        return True

    def _generate_random_scene(self, uuid=0):
        """
        Draw random samples in random positions
        """
        bg = Image.open(self.background_path)
        bg_w, bg_h = bg.size
        perc_area = random.random()
        bg_area = bg_w * bg_h * perc_area
        samples = []
        while bg_area > 0.0:
            # Select random category name
            class_name_lst = list(self.class_samples.keys())
            class_name = random.choice(class_name_lst)
            class_id = self.class_idx[class_name]
            if not self.class_samples[class_name]:
                click.echo("There are not samples in %s" % class_name)
                continue
            sample_path = random.choice(self.class_samples[class_name])
            img = Image.open(sample_path)
            img_w, img_h = img.size
            if self.sample_perc_size != 1.0:
                img_w = math.floor(img_w * self.sample_perc_size)
                img_h = math.floor(img_h * self.sample_perc_size)
                img.thumbnail((img_w, img_h))
            if self.sample_degree != 0:
                angle = random.randint(0, self.sample_degree)
                img = img.rotate(angle, expand=True)
                img_w, img_h = img.size
            if img_w * img_h > bg_w * bg_h:
                click.echo(
                    "sample %s is greater than background" % sample_path
                )
                continue
            bg_area -= img_w * img_h
            pos_x = random.randint(-img_w // 2, bg_w - img_w // 2)
            pos_y = random.randint(-img_h // 2, bg_h - img_h // 2)
            samples.append(
                Sample(class_id, img, pos_x, pos_y,
                       bg_w, bg_h, len(class_name_lst))
            )
            # bg.paste(img, (pos_x, pos_y))
            # # TODO: if bg is .png:
            # bg = Image.alpha_composite(
            #     Image.new("RGBA", bg.size),
            #     bg.convert('RGBA')
            # )
            bg.paste(img, (pos_x, pos_y), img)

        # Write File
        file_path = os.path.join(self.output_path, 'frame_%s.txt' % str(uuid))
        f = open(file_path, 'w+')
        f.writelines([' '.join(x.get_yolo_format()) + '\n' for x in samples])
        f.close()
        # Save Image
        img_path = os.path.join(self.output_path, 'frame_%s.jpg' % str(uuid))
        bg.save(img_path)
        return True

    def generate_random_scenes(self, num_frames=1):
        if not self.enable_generator:
            return True
        files_qty = len(glob(self.output_path + '/*.jpg'))
        with click.progressbar(range(0, num_frames),
                               label='Creating Stages') as bar:
            for i in bar:
                click.echo(
                    "Creating Random scene %s of %s" % (i + 1, num_frames)
                )
                self._generate_random_scene(files_qty + i)
        return True



# -*- coding: utf-8 -*-
###############################################################################
#   Copyright (c) 2021 Gaston Alberto Bertolani
#   License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
###############################################################################

import numpy as np
import cv2
from PIL import Image
from operator import xor
import os
import re
from glob import glob


class BackgroundEliminator(object):

    def __init__(self, image_path, save_path=False,
                 no_origin=False, with_groups=False):
        self.image_path = image_path
        self.save_path = save_path
        self.isfile = os.path.isfile(self.image_path)
        self.no_origin = no_origin
        self.with_groups = with_groups

    def _get_img_save_path(self, image_path):
        res = re.search('.*\.(\w{3,4})$', image_path)
        ft = res.group(1)
        if not self.save_path:
            save_path = image_path[:-len(ft)] + 'png'
        elif os.path.isdir(self.save_path):
            # If it is with a group, we create the group
            # of the image in the destination
            dir_to_save = self.save_path
            if self.with_groups:
                base_path = os.path.dirname(image_path)
                category_name = os.path.basename(base_path)
                dir_to_save = os.path.join(self.save_path, category_name)
                try:
                    os.mkdir(dir_to_save)
                except FileExistsError:
                    pass
            image_name = os.path.basename(image_path)[:-len(ft)] + 'png'
            save_path = os.path.join(dir_to_save, image_name)
        else:
            save_path = self.save_path
        return save_path

    def _navigate(self, folder):
        folders = glob(folder + '/*/')
        # Folder is last folder
        if not folders:
            image_path_lst = []
            image_formats = ('*.png', '*.jpg', '*.jpeg', '*.bmp')
            for format in image_formats:
                image_path_lst.extend(glob(folder + '/' + format))
            for i, image_path in enumerate(image_path_lst):
                click.echo("Removing background %s of %s"
                      % (i+1, len(image_path_lst)))
                self._remove_background(image_path)
        else:
            for folder in folders:
                self._navigate(folder)

    def _remove_background(self, image_path):
        img = cv2.imread(image_path)
        # convert to graky
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        # threshold input image as mask
        mask = cv2.threshold(gray, 250, 255, cv2.THRESH_BINARY)[1]

        # negate mask
        mask = 255 - mask

        # apply morphology to remove isolated extraneous noise
        # use borderconstant of black since foreground touches the edges
        kernel = np.ones((3, 3), np.uint8)
        mask = cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel)
        mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, kernel)

        # anti-alias the mask -- blur then stretch
        # blur alpha channel
        mask = cv2.GaussianBlur(mask, (0, 0), sigmaX=2, sigmaY=2,
                                borderType=cv2.BORDER_DEFAULT)

        # linear stretch so that 127.5 goes to 0, but 255 stays 255
        mask = (2*(mask.astype(np.float32))-255.0).clip(0, 255).astype(np.uint8)

        # put mask into alpha channel
        result = img.copy()
        result = cv2.cvtColor(result, cv2.COLOR_BGR2BGRA)
        result[:, :, 3] = mask

        save_path = self._get_img_save_path(image_path)

        cv2.imwrite(save_path, result)

        img = np.array(Image.open(save_path))
        idx = np.where(img[:, :, 3] > 0)
        x0, y0, x1, y1 = idx[1].min(), idx[0].min(), idx[1].max(), idx[0].max()
        out = Image.fromarray(img[y0:y1+1, x0:x1+1, :])
        out.save(save_path)
        if self.no_origin:
            os.remove(image_path)
        return True

    def remove_background(self):
        if self.isfile:
            self._remove_background(self.image_path)
        else:
            self._navigate(self.image_path)
        return True



# -*- coding: utf-8 -*-
###############################################################################
#   Copyright (c) 2021 Gaston Alberto Bertolani
#   License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
###############################################################################


import random
from shutil import copyfile
from pathlib import Path
from glob import glob
import os

import click


class Separator(object):

    """
    Separate files in train, test and validation folder
    accoirding to the percentages entered
    """

    def __init__(self, samples_path, train_path,
                 validation_path=False, test_path=False,
                 train_perc=0.0, validation_perc=0.0,
                 test_perc=0.0):
        if not train_path and \
                not validation_path and not test_path:
            raise click.ClickException("Target Path is required")
        if not train_perc and \
                not validation_perc and not test_perc:
            raise click.ClickException("Target Perc is required")
        if validation_perc != 0.0 and not validation_path:
            raise click.ClickException("Validation path is required")
        if test_perc != 0.0 and not test_path:
            raise click.ClickException("Test path is required")
        if train_perc < 0.0 or \
                test_perc < 0.0 or validation_perc < 0.0:
            raise click.ClickException("Percentage must be positive")
        total_perc = train_perc + test_perc + validation_perc
        if total_perc != 100.0:
            raise click.ClickException("Total percentage must be 100%%")
        self.samples_path = samples_path
        self.train_path = train_path
        self.validation_path = validation_path
        self.test_path = test_path
        self.train_perc = train_perc
        self.validation_perc = validation_perc
        self.test_perc = test_perc

    def get_target_path(self, percent):
        perc_lst = [
            (self.train_path, self.train_perc),
        ]
        if self.validation_path:
            perc_lst.append((self.validation_path, self.validation_perc))
        if self.test_path:
            perc_lst.append((self.test_path, self.test_perc))
        perc_lst.sort(key=lambda x: x[1], reverse=True)
        accu = 0.0
        for perc_el in perc_lst:
            accu += (perc_el[1] / 100)
            if percent <= accu:
                return perc_el[0]

    def split_samples(self, format=False, remove_source=False):
        if isinstance(format, str):
            format = format.split(',')
        if (not isinstance(format, list) or
                not isinstance(format, tuple)) and format:
            raise click.ClickException("Invalid Format %s." % format)
        if not format:
            # Get all files
            format = ['*.*']
        map_files = {}
        for ft in format:
            if '*.' not in ft:
                ft = '*.' + ft
            for file_path in glob("%s/%s" % (self.samples_path, ft)):
                fpath = Path(file_path)
                if fpath.stem not in map_files:
                    map_files[fpath.stem] = []
                map_files[fpath.stem].append(fpath.suffix)
        # Random selection and random split
        file_name_lst = list(map_files.keys())
        for i in range(0, len(file_name_lst)):
            file_name = random.choice(file_name_lst)
            file_name_lst.remove(file_name)
            perc = random.random()  # [0, 1]
            folder_path = self.get_target_path(perc)
            for ft in map_files[file_name]:
                copyfile(
                    os.path.join(self.samples_path, file_name + ft),
                    os.path.join(folder_path, file_name + ft),
                )
        click.secho("Done!", fg='green', bold=True, blink=True)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

install_requires=[
    'cycler',
    'kiwisolver',
    'matplotlib',
    'numpy',
    'opencv-python-headless',
    'Pillow',
    'pyparsing',
    'python-dateutil',
    'six',
    'click',
]

setup(
    name='ai-cv-utils',
    version='0.0.3',
    author='Gaston Alberto Bertolani',
    author_email='gbertolani.in@gmail.com',
    description='AI Computer Vision Utilities',
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    url='https://github.com/gbertolani/ai-cv-utils',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Utilities',

    ],
    license='AGPL3+',
    packages=find_packages(),
    install_requires=install_requires,
    python_requires='>=3.6',
    entry_points='''
        [console_scripts]
        aicv=aicv.main:aicv
    ''',

)
